#!/usr/bin/env python3

# mceliece parameters:

m = 13
n = 6960
t = 119

# mctiny parameters:

x = 680
y = 13
v = 7

# mceliece parameter requirements:

assert m >= 1
assert n >= 1
assert t >= 2
assert n <= 2**m

# mctiny parameter requirements:

assert x > 0
assert y > 0
assert v > 0
assert x%8 == 0

# derived parameters:

rows = t*m # number of rows in public key
cols = n-rows # number of columns in public key
rowbytes = (cols+7)//8 # number of bytes in one row
colbytes = (rows+7)//8 # number of bytes in one column
ebytes = (n+7)//8 # number of bytes in error vector

xbytes = (x+7)//8
colblocks = (cols+x-1)//x
ybytes = (y+7)//8
rowblocks = (rows+y-1)//y
blockbytes = xbytes*y

pieces = (rows+v*y-1)//(v*y) # separately delivered pieces of syndrome
piecebytes = (v*y+7)//8 # bytes per piece of syndrome; final piece may be shorter

ciphertextbytes = colbytes+32 # 32-byte plaintext-confirmation hash

# cookie0 holds key to the universe:
# 16-byte authenticator, 32-byte kmaster, 32-byte seed, 1-byte cookie-key id
cookie0bytes = 16+32+32+1

# cookieblock holds one block's contribution to ciphertext:
# 16-byte authenticator, y-bit synd1, 1-byte cookie-key id
cookieblockbytes = 16+ybytes+1

# cookie9 holds session key:
# 16-byte authenticator, 32-byte ksession, 1-byte cookie-key id
cookie9bytes = 16+32+1

# overhead in each packet:
# 16-byte authenticator, 24-byte nonce
universal = 16+24

query0bytes = universal+ciphertextbytes+32+512 # 32-byte serverpkhash, 512-byte extensions
reply0bytes = universal+cookie0bytes
query1bytes = universal+cookie0bytes+blockbytes
reply1bytes = universal+cookie0bytes+cookieblockbytes
query2bytes = universal+cookie0bytes+v*colblocks*cookieblockbytes
reply2bytes = universal+cookie0bytes+piecebytes
query3bytes = universal+cookie0bytes+colbytes
reply3bytes = universal+cookie9bytes+ciphertextbytes

# derived parameter requirements:

assert colblocks <= 32
assert rowblocks <= 128

assert query0bytes <= 1252
assert reply0bytes <= 1252
assert query1bytes <= 1252
assert reply1bytes <= 1252
assert query2bytes <= 1252
assert reply2bytes <= 1252
assert query3bytes <= 1252
assert reply3bytes <= 1252

# outputs:

print('/* See mctiny.md for documentation. */')
print('')
print('/* For implementors: This file is auto-generated by mctiny.py; do not edit. */')
print('/* See mctiny.py for the underlying calculations. */')
print('')
print('#ifndef mctiny_h')
print('#define mctiny_h')
print('')
print('#include "crypto_kem_mceliece%d%d.h"' % (n,t))
print('')
print('#define mctiny_PUBLICKEYBYTES crypto_kem_mceliece%d%d_PUBLICKEYBYTES' % (n,t))
print('#define mctiny_SECRETKEYBYTES crypto_kem_mceliece%d%d_SECRETKEYBYTES' % (n,t))
print('#define mctiny_CIPHERTEXTBYTES crypto_kem_mceliece%d%d_CIPHERTEXTBYTES' % (n,t))
print('#define mctiny_SESSIONKEYBYTES crypto_kem_mceliece%d%d_BYTES' % (n,t))
print('')
print('#define mctiny_keypair crypto_kem_mceliece%d%d_keypair' % (n,t))
print('#define mctiny_enc crypto_kem_mceliece%d%d_enc' % (n,t))
print('#define mctiny_dec crypto_kem_mceliece%d%d_dec' % (n,t))
print('')
print('#define mctiny_M %d' % m)
print('#define mctiny_MMASK %d' % ((2**m)-1))
print('#define mctiny_N %d' % n)
print('#define mctiny_T %d' % t)
print('#define mctiny_ROWBITS %d' % cols)
print('#define mctiny_ROWBYTES %d' % rowbytes)
print('#define mctiny_COLBITS %d' % rows)
print('#define mctiny_COLBYTES %d' % colbytes)
print('')
print('#define mctiny_EBYTES %d' % ebytes)
print('')
print('#define mctiny_X %d' % x)
print('#define mctiny_XBYTES %d' % xbytes)
print('#define mctiny_COLBLOCKS %d' % colblocks)
print('')
print('#define mctiny_Y %d' % y)
print('#define mctiny_YBYTES %d' % ybytes)
print('#define mctiny_ROWBLOCKS %d' % rowblocks)
print('#define mctiny_BLOCKBYTES %d' % blockbytes)
print('')
print('#define mctiny_V %d' % v)
print('')
print('#define mctiny_PIECES %d' % pieces)
print('')
print('#define mctiny_PIECEBYTES %d' % piecebytes)
print('')
print('#define mctiny_COOKIE0BYTES %d' % cookie0bytes)
print('#define mctiny_COOKIEBLOCKBYTES %d' % cookieblockbytes)
print('#define mctiny_COOKIE9BYTES %d' % cookie9bytes)
print('')
print('#define mctiny_QUERY0BYTES %d' % query0bytes)
print('#define mctiny_REPLY0BYTES %d' % reply0bytes)
print('#define mctiny_QUERY1BYTES %d' % query1bytes)
print('#define mctiny_REPLY1BYTES %d' % reply1bytes)
print('#define mctiny_QUERY2BYTES %d' % query2bytes)
print('#define mctiny_REPLY2BYTES %d' % reply2bytes)
print('#define mctiny_QUERY3BYTES %d' % query3bytes)
print('#define mctiny_REPLY3BYTES %d' % reply3bytes)
print('')
print('extern int mctiny_seedisvalid(const unsigned char *);')
print('extern void mctiny_seed2e(unsigned char *,const unsigned char *);')
print('')
print('extern void mctiny_pk2block(unsigned char *,const unsigned char *,int,int);')
print('extern void mctiny_eblock2syndrome(unsigned char *,const unsigned char *,const unsigned char *,int);')
print('')
print('extern void mctiny_pieceinit(unsigned char *,const unsigned char *,int);')
print('extern void mctiny_pieceabsorb(unsigned char *,const unsigned char *,int);')
print('')
print('extern void mctiny_mergepieces(unsigned char *,const unsigned char (*)[mctiny_PIECEBYTES]);')
print('')
print('extern void mctiny_finalize(unsigned char *, unsigned char *,const unsigned char *,const unsigned char *);')
print('')
print('#endif')
